# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-10-11 13:24:56
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-06-30 14:01:05

from __future__ import absolute_import, division, print_function

import os
import sys
import glob
import re
from collections import OrderedDict
import six
import json
from tree import log, config as cfg_params

if ((sys.version_info.major == 3 and sys.version_info.minor > 2) or
        (sys.version_info.major == 2 and sys.version_info.minor >= 7)):
    from configparser import ConfigParser as SafeConfigParser
else:
    from configparser import SafeConfigParser

orig_environ = os.environ.copy()


class Tree(object):
    ''' Initialize the sdss tree object

    This class provides Python programmatic access to the SDSS tree envionment structure

    Parameters:
        key (str|list):
            A section or list of sections of the tree to add into the local environment
        uproot_with (str):
            A new TREE_DIR path used to override an existing TREE_DIR environment variable
        config (str):
            Name of manual config file to load.  Default is sdsswork.
        update (bool):
            If True, overwrites existing tree environment variables in your
            local environment.  Default is False.
        exclude (list):
            A list of environment variables to exclude
            from forced updates
        root (str):
            An absolute directory path to override as the software product root
        git (bool):
            If True, looks for SDSS_GIT_ROOT environment variable as product root instead
            of SDSS_SVN_ROOT

    Attributes:
        treedir (str):
            The directory of the tree
        environ (dict):
            The fully loaded SDSS config file held internally

    '''
    # possible software product roots
    _product_roots = ['PRODUCT_ROOT', 'SDSS_GIT_ROOT', 'SDSS_SVN_ROOT', 'SDSS_INSTALL_PRODUCT_ROOT',
                      'SDSS_PRODUCT_ROOT', 'SDSS4_PRODUCT_ROOT']

    def __init__(self, config=None, key=None, uproot_with=None, update=None, exclude=None,
                 product_root=None, git=None):
        self.config_name = config or 'sdsswork'
        self.exclude = exclude or []
        update = update or False
        self._keys = key
        self._file_replace = '@FILESYSTEM@'

        # set the roots
        self.set_roots(uproot_with=uproot_with)

        # load the configuraiton file
        self.load_config()

        # create the environment
        self.branch_out(limb=key)

        # add the paths to the os.environ
        if key is not None:
            self.add_paths_to_os(key='general', update=update)
        self.add_paths_to_os(key=key, update=update)

        # set the software product root, $PRODUCT_ROOT envvar
        self.productroot_dir = None
        self.set_product_root(root=product_root, git=git)

    def __repr__(self):
        return ('Tree(sas_base_dir={0}, config={1})'.format(self.sasbasedir, self.config_name))

    @property
    def phase(self):
        ''' Return the phase of the survey from the loaded "work" environment '''
        phase = self.environ['default'].get('phase', None)
        if phase and phase.isdigit():
            phase = int(phase)
        return phase

    def list_keys(self):
        ''' List the available keys you can load '''
        return [k for k in self.environ.keys() if k not in ['general', 'default']]

    def set_roots(self, uproot_with=None):
        ''' Set the roots of the tree in the os environment

        Parameters:
            uproot_with (str):
                A new TREE_DIR path used to override an existing TREE_DIR environment variable

        '''

        # Check for TREE_DIR
        self.treedir = get_tree_dir(uproot_with=uproot_with)

        # Check sas_base_dir
        if 'SAS_BASE_DIR' in os.environ:
            self.sasbasedir = os.environ["SAS_BASE_DIR"]
        else:
            self.sasbasedir = os.path.expanduser('~/sas')

        # make the directories
        if not os.path.isdir(self.sasbasedir):
            os.makedirs(self.sasbasedir)

    def _read_config(self, config=None, bases=None):
        ''' Read a config file

        Uses ConfigParser to read in a cfg file.  If a base
        is identified, then recursively reads in all cfg files and builds
        a master config dictionary object

        Parameters:
            config (str):
                Optional name of config file to load
            bases (list):
                A list of parent config files

        Returns:
            A configParser dictionary object
        '''

        # format name and file
        config_name = config if config.endswith('.cfg') else '{0}.cfg'.format(config)
        config_file = os.path.join(self.treedir, 'data', config_name)
        assert os.path.isfile(
            config_file) is True, 'config file {0} must exist in the data directory'.format(config_file)

        # read initial config file
        cfg = SafeConfigParser()
        cfg.optionxform = lambda option: option
        cfg.read(config_file)

        # check for any bases
        bases = bases if bases else []
        bases.insert(0, os.path.join(self.treedir, 'data', config_name))
        hasbase = 'base' in cfg.defaults()

        # read base config file
        if hasbase:
            return self._read_config(cfg.defaults()['base'], bases=bases)
        else:
            # read in the full list of config files
            cfg = SafeConfigParser()
            cfg.optionxform = lambda option: option
            cfg.read(bases)
            # if both eboss and boss, then remove boss
            if 'EBOSS' in cfg.sections() and 'BOSS' in cfg.sections():
                cfg.remove_section('BOSS')
        return cfg

    def _check_config(self, config=None):
        ''' check the config

        checks the config for syntax and existence.  Defaults to sdsswork
        or a DR if it doesn't exist.

        Parameters:
            config (str):
                The name of the config to check

        Returns:
            The (updated) name of the config
        '''

        # check initial argument
        cfgname = (config or self.config_name)
        cfgname = 'sdsswork' if cfgname is None else cfgname
        assert isinstance(cfgname, six.string_types), 'config name must be a string'
        cfgname = cfgname.lower()

        config_name = cfgname if cfgname.endswith('.cfg') else '{0}.cfg'.format(cfgname)

        # default to another config if not available
        file_base = [os.path.basename(f) for f in self._cfg_files]
        if config_name not in file_base:
            if 'dr' in config_name:
                drint = max(map(int, [re.findall(r'\d{1,2}', f)[0] for f in file_base if 'dr' in f]))
                log.warning('{0} not found. Defaulting to dr{1}.cfg'.format(config_name, drint))
                self.config_name = 'dr{0}'.format(drint)
            else:
                log.warning('{0} not found. Defaulting to sdsswork'.format(config_name))
                self.config_name = 'sdsswork'
            config_name = self.config_name + '.cfg'
        return config_name

    def load_config(self, config=None):
        ''' Load a config file

        Parameters:
            config (str):
                Optional name of manual config file to load
        '''

        # get a list of all config files
        self._cfg_files = glob.glob(os.path.join(self.treedir, 'data', '*.cfg'))
        self._cfg_files.sort()

        # check the config file
        config_name = self._check_config(config)

        # set the confifile
        configfile = os.path.join(self.treedir, 'data', config_name)
        assert os.path.isfile(configfile) is True, ('configfile {0} must exist in the '
                                                    'data directory'.format(configfile))
        self.config_file = configfile

        # read the config
        self._cfg = self._read_config(config=self.config_name)

    def _create_environment(self, cfg=None, sections=None):
        ''' create the environment from the config

        Creates a dictionary with environment definitions
        expanded out

        Parameters:
            config (str):
                Optional name of manual config file to load
            sections (list):
                A list of config sections to load

        Returns:
            An ordered dictionary of envvar definitions
        '''

        # pass in a cfg dict or use the one attached
        cfg = cfg or self._cfg

        # create the local tree environment
        environ = OrderedDict()
        environ['default'] = cfg.defaults()

        # set the filesystem envvar to sas_base_dir
        filesystem = 'FILESYSTEM' if 'FILESYSTEM' in environ[
            'default'] else 'filesystem' if 'filesystem' in environ['default'] else None
        if environ['default'][filesystem] == self._file_replace:
            environ['default'][filesystem] = self.sasbasedir

        # add all sections into the tree environ
        sections = sections if sections else cfg.sections()
        for section in sections:
            # skip if PATHS
            if section == 'PATHS':
                continue

            section = section if section in cfg.sections() else section.upper()
            environ[section] = OrderedDict()
            options = cfg.options(section)

            for opt in options:
                if opt in environ['default']:
                    continue
                val = cfg.get(section, opt)
                if val.find(self._file_replace) == 0:
                    val = val.replace(self._file_replace, self.sasbasedir)
                environ[section][opt] = val

        return environ

    def _create_paths(self, cfg=None):
        ''' create a dictionary of path definitions

        Extracts the PATHS section from a ConfigParser object
        and builds a dictionary of paths for sdss_access

        Parameters:
            cfg (object):
                A configParser object
        Returns:
            An ordered dictionary of sdss_access path definitions
        '''

        # pass in a cfg dict or use the one attached
        cfg = cfg or self._cfg

        # return if no PATHS found
        if not cfg.has_section('PATHS'):
            return None

        paths = OrderedDict()
        for opt in cfg.options('PATHS'):
            if opt in cfg.defaults():
                continue
            paths[opt] = cfg.get('PATHS', opt)

        # sort the paths by name
        paths = OrderedDict({k: paths[k] for k in sorted(paths.keys())})
        return paths

    def branch_out(self, limb=None):
        ''' Set the individual section branches

        This adds the various sections of the config file into the
        tree environment for access later. Optionally can specify a specific
        branch.  This does not yet load them into the os environment.

        Parameters:
            limb (str|list):
                A section or lists of sections of the config to add into the environ

        '''

        # Filter on sections
        if not limb:
            limbs = self._cfg.sections()
        else:
            # we must have the general always + secton
            limb = limb if isinstance(limb, list) else [limb]
            limbs = ['general']
            limbs.extend(limb)

        # add all limbs into the tree environ
        self.environ = self._create_environment(sections=limbs)

        # add all paths into the tree paths dictionary
        self.paths = self._create_paths()

    def add_limbs(self, key=None):
        ''' Add a new section from the tree into the existing os environment

        Parameters:
            key (str):
                The section name to grab from the environment

        '''
        self.branch_out(limb=key)
        self.add_paths_to_os(key=key)

    def get_paths(self, key):
        ''' Retrieve a set of environment paths from the config

        Parameters:
            key (str):
                The section name to grab from the environment

        Returns:
            self.environ[newkey] (OrderedDict):
                An ordered dict containing all of the paths from the
                specified section, as key:val = name:path
        '''
        newkey = key if key in self.environ else key.upper() if key.upper() \
            in self.environ else None
        if newkey:
            return self.environ[newkey]
        else:
            raise KeyError('Key {0} not found in tree environment'.format(key))

    def add_paths_to_os(self, key=None, update=None):
        ''' Add the paths in tree environ into the os environ

        This code goes through the tree environ and checks
        for existence in the os environ, then adds them

        Parameters:
            key (str):
                The section name to check against / add
            update (bool):
                If True, overwrites existing tree environment variables in your
                local environment.  Default is False.
        '''

        if key is not None:
            allpaths = key if isinstance(key, list) else [key]
        else:
            allpaths = [k for k in self.environ.keys() if 'default' not in k]

        for key in allpaths:
            paths = self.get_paths(key)
            self._check_paths(paths, update=update)

    def _check_paths(self, paths, update=None):
        ''' Check if the path is in the os environ, and if not add it

        Paramters:
            paths (OrderedDict):
                An ordered dict containing all of the paths from the
                a given section, as key:val = name:path
            update (bool):
                If True, overwrites existing tree environment variables in your
                local environment.  Default is False.
        '''

        # set up the exclusion list
        exclude = [] if not self.exclude else self.exclude \
            if isinstance(self.exclude, list) else [self.exclude]

        # check the path names
        for pathname, path in paths.items():
            if update and pathname.upper() not in exclude:
                os.environ[pathname.upper()] = os.path.normpath(path)
            elif pathname.upper() not in os.environ:
                os.environ[pathname.upper()] = os.path.normpath(path)

    def replant_tree(self, config=None, exclude=None):
        ''' Replant the tree with a different config setup

        Parameters:
            config (str):
                The config name to reload
            exclude (list):
                A list of environment variables to exclude
                from forced updates
        '''

        # reinitialize a new Tree with a new config
        config = 'sdsswork' if not config else config
        self.__init__(key=self._keys, config=config, update=True, exclude=exclude)

    def list_configs(self):
        ''' List available configs to load '''

        files = [os.path.basename(f) for f in self._cfg_files if 'basework.cfg' not in f]
        return files

    def show_forest(self, config=None):
        ''' Show the environment for a specified config

        Creates a dictionary environment for each config in the list of
        available configurations.

        Parameters:
            config (str):
                The config to show

        Returns:
            A dictionary of config environment(s)
        '''

        configs = [config] if config else self.list_configs()
        cfgs = {}
        for cfg in configs:
            cfg_name = cfg.split('.cfg')[0]
            cfg = self._read_config(cfg_name)
            cfgs[cfg_name] = self._create_environment(cfg)

        if len(configs) == 1:
            return cfgs[cfg_name]
        return cfgs

    @staticmethod
    def list_available_configs():
        ''' List the available config files able to be loaded '''

        # get tree dir
        treedir = get_tree_dir()

        # look up the config files from the data directory
        data_path = os.path.join(treedir, 'data')
        cfgs = [i for i in os.listdir(data_path) if i.endswith('.cfg')]
        cfgs.sort()

        # sort the DR subset of config files
        drsort = sorted([i for i in cfgs if 'dr' in i],
                        key=lambda t: int(re.findall('dr(.*?).cfg', t)[0]))
        rest = [[i] for i in cfgs if 'dr' not in i]
        rest.insert(1, drsort)
        sorted_cfgs = sum(rest, [])
        return sorted_cfgs

    @classmethod
    def get_available_releases(cls, public=None):
        ''' Get the available releases

        Parameters:
            public (bool):
                If True, only return public data releases
        '''

        # get the configs
        cfgs = cls.list_available_configs()

        # parse the data releases
        releases = []
        for i in cfgs:
            if public and not i.startswith('dr'):
                continue

            b = i.split('.cfg', 1)[0]
            if 'dr' in b or 'mpl' in b:
                # uppercase any DR or MPLs
                releases.append(b.upper())
            elif 'work' in b and 'WORK' not in releases:
                # reduce alll xxxxwork cfgs to a single "work" release
                releases.append('WORK')
        return releases

    @staticmethod
    def reset_os_environ():
        ''' Resets os.environ with the orignal cache before tree mods '''
        os.environ = orig_environ

    @staticmethod
    def get_orig_os_environ():
        ''' Returns the original os.environ '''
        return orig_environ

    def to_dict(self, collapse=True):
        ''' Convert tree environment to standard dicts

        Converts the nested ``tree.environ`` into a series of
        ordinary dicts.

        Parameters:
            collapse (bool):
                If True, collapses nested dicts into a single dict.  Default is True.
        '''
        if collapse is False:
            dd = json.loads(json.dumps(self.environ))
            dd.pop('default')
            return dd

        dd = {}
        for k, v in self.environ.items():
            if k == 'default':
                continue
            for kk, vv in v.items():
                dd[kk] = vv
        return dd

    @staticmethod
    def get_product_root(root=None, git=None):
        ''' Get the sdss product root used for svn/git products

        Attempts to extract the root directory for SDSS-installed git/svn products.
        Uses the following environment variables in order of precendence:
        PRODUCT_ROOT, SDSS_SVN_ROOT, SDSS_INSTALL_PRODUCT_ROOT, SDSS_PRODUCT_ROOT,
        SDSS4_PRODUCT_ROOT. If no root is found uses one directory up from SAS_BASE_DIR.

        Parameters:
            root (str):
                An absolute directory path to override as the product root
            git (bool):
                If True, looks for SDSS_GIT_ROOT environment variable as product root.

        Returns:
            The directory path to sdss-installed svn/git products
        '''

        # override with an input root directory
        if root:
            return root

        # use an existing $PRODUCT_ROOT envvar
        product_root = os.getenv("PRODUCT_ROOT", None)
        if product_root:
            return product_root

        # check in the config for a product root definition
        product_root = cfg_params.get("PRODUCT_ROOT", None)
        if product_root:
            return product_root

        # attempt to extract a product root from a variety of environment variables
        repo_root = 'SDSS_GIT_ROOT' if git else 'SDSS_SVN_ROOT'
        product_root = os.getenv(repo_root, os.getenv(
            "SDSS_INSTALL_PRODUCT_ROOT", os.getenv("SDSS4_PRODUCT_ROOT", None)))
        if not product_root:
            product_root = os.getenv("SAS_BASE_DIR").rsplit('/', 1)[0]

        return product_root

    def set_product_root(self, root=None, git=None):
        ''' Sets the sdss product root used for svn/git products

        Sets the root directory for SDSS-installed git/svn products as
        the $PRODUCT_ROOT environment variable.  Attempts to find a viable $PRODUCT_ROOT
        using ``get_product_method``.  Viable product roots in order of precendence:
        SDSS_SVN_ROOT, SDSS_INSTALL_PRODUCT_ROOT, SDSS_PRODUCT_ROOT, SDSS4_PRODUCT_ROOT.
        If no root is found uses one directory up from SAS_BASE_DIR.

        Parameters:
            root (str):
                An absolute directory path to override as the product root
            git (bool):
                If True, looks for SDSS_GIT_ROOT environment variable as product root.

        '''
        product_root = self.get_product_root(root=root, git=git)
        os.environ['PRODUCT_ROOT'] = product_root
        self.productroot_dir = product_root

    def write_old_paths_inifile(self):
        ''' Write out an old sdss_paths ini file '''

        paths_dir = os.path.join(self.treedir, 'data/sdss_paths.ini')
        with open(paths_dir, 'w') as f:
            f.write('# Paths for SDSS files. Each file type is given a template full path.\n')
            f.write('# \n')
            f.write('# This file has been deprecated as of tree/3.0.0 and sdss_access/1.0.0 \n')
            f.write('# It should not be updated manually.  Use the tree.write_olds_paths method in \n')
            f.write('# the python tree code to generate a new version of this file as needed.\n')
            f.write('# \n')
            f.write('\n')
            f.write("[paths]\n")
            for name, template in self.paths.items():
                # switch special functions back to %
                template = template.replace('@', '%')
                # write out path, template
                f.write('{0} = {1}\n'.format(name, template))


def get_tree_dir(uproot_with=None):
    ''' Return the path to the tree product directory

    Parameters:
        uproot_with (str):
            A new TREE_DIR path used to override an existing TREE_DIR environment variable

    Returns:
        The path to the tree python product directory
    '''
    treedir = os.environ.get('TREE_DIR', None) if not uproot_with else uproot_with
    if not treedir:
        treefilepath = os.path.dirname(os.path.abspath(__file__))
        if 'python/' in treefilepath:
            treedir = treefilepath.rsplit('/', 2)[0]
        else:
            treedir = treefilepath
        treedir = treefilepath
        os.environ['TREE_DIR'] = treedir
    return treedir



# encoding: utf-8
#
# setup.py
#


import distutils.dir_util
from setuptools import setup

# Copy data files into the python/tree/data directory
tmp = distutils.dir_util.copy_tree('data', 'python/tree/data')


setup()




# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-08 16:31:45
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-31 15:39:47

from __future__ import print_function, division, absolute_import
import os
import re
import pytest
from sdss_access import tree
from sdss_access.path import Path
from tests.conftest import gzcompress, gzuncompress


class TestPath(object):

    def test_missing_keys(self, path):
        with pytest.raises(KeyError) as cm:
            full = path.full('mangacube')
        assert 'Missing required keyword arguments:' in str(cm.value)

    def test_public(self):
        path = Path()
        url = path.url('mangacube', drpver='v2_3_1', plate=8485, ifu='1901', wave='LOG')
        assert 'mangawork' in url

        release = 'dr14'
        path = Path(public=True, release=release)
        url = path.url('mangacube', drpver='v2_3_1', plate=8485, ifu='1901', wave='LOG')
        assert release in url

    @pytest.mark.parametrize('place, exp', [('local', False), ('remote', True)])
    def test_existence(self, path, place, exp):
        full = path.full('mangaimage', drpver='v2_5_3', plate=8116, ifu=1901, dir3d='mastar')
        exists = path.exists('', full=full, remote=(place == 'remote'))
        assert exp == exists

    def test_lookup_names(self, path):
        assert 'mangacube' in path.lookup_names()

    @pytest.mark.parametrize('name, keys', [('mangacube', ['drpver', 'ifu', 'plate', 'wave']),
                                            ('plateLines', ['plateid'])])
    def test_lookup_keys(self, path, name, keys):
        realkeys = path.lookup_keys(name)
        assert set(keys) == set(realkeys)

    @pytest.mark.parametrize('name, special, keys, exp',
                             [('plateLines', '@platedir', {'plateid': 8485},
                               '0084XX/008485/plateLines-008485.png'),
                              ('plateLines', '@platedir', {'plateid': 11002},
                               '0110XX/011002/plateLines-11002.png'),
                              ('spAll', '@spectrodir', {'run2d': 103},
                               'sas/sdsswork/sdss/spectro'),
                              ('spAll', '@spectrodir', {'run2d': 100},
                               'sas/ebosswork/eboss/spectro')],
                             ids=['platedir_p4', 'platedir_p5', 'spectrodir_r1', 'spectrodir_r2'])
    def test_special_function(self, path, name, special, keys, exp):
        assert special in path.templates[name]
        full = path.full(name, **keys)
        assert exp in full

    def test_envvar_expansion(self, path):
        name = 'mangacube'
        assert '$MANGA_SPECTRO_REDUX' in path.templates[name]
        full = path.full(name, drpver='v2_4_3', plate=8485, ifu=1901, wave='LOG')
        exp = 'sas/mangawork/manga/spectro/redux/v2_4_3/8485'
        assert exp in full

    @pytest.mark.parametrize('name, example, keys',
                             [('mangacube',
                               'mangawork/manga/spectro/redux/v2_4_3/8485/stack/manga-8485-1901-LOGCUBE.fits.gz',
                               {'drpver': 'v2_4_3', 'plate': '8485', 'ifu': '1901', 'wave': 'LOG'}),
                              ('REJECT_MASK', 'ebosswork/eboss/lss/reject_mask/mask.html',
                               {'type': 'mask', 'format': 'html'}),
                              ('fpBIN', 'ebosswork/eboss/photo/redux/1/45/objcs/3/fpBIN-000045-g3-0123.fit',
                               {'rerun': '1', 'field': '0123', 'filter': 'g', 'camcol': '3', 'run': '000045'}),
                              ('galaxy', 'ebosswork/eboss/lss/galaxy_DR12v1.0_1_n.fits.gz',
                               {'sample': '1', 'dr': 'DR12', 'version': 'v1.0', 'ns': 'n'})],
                             ids=['mangacube', 'reject', 'fpbin', 'galaxy'])
    def test_extract(self, path, name, example, keys):
        fullpath = os.path.join(os.environ['SAS_BASE_DIR'], example)
        realkeys = path.extract(name, fullpath)
        assert keys == realkeys

    def test_extract_source(self, path):
        code = path._find_source(path.full)
        assert 'def full(self' in code
        assert 'template = self._call_special_functions' in code

    def full(self, path, name='mangacube'):
        full = path.full(name, drpver='v2_4_3', plate=8485, ifu='*', wave='LOG')
        return full

    def test_location(self, path):
        full = self.full(path)
        loc = path.location('', full=full)
        assert 'mangawork/manga/spectro/redux/v2_4_3/8485/stack/manga-8485-*-LOGCUBE.fits.gz' in loc

    def test_name(self, path):
        full = self.full(path)
        name = path.name('', full=full)
        assert 'manga-8485-*-LOGCUBE.fits.gz' in name

    def test_dir(self, path):
        full = self.full(path)
        d = path.dir('', full=full)
        assert d.endswith('mangawork/manga/spectro/redux/v2_4_3/8485/stack')

    def test_url(self, path):
        full = self.full(path)
        url = path.url('', full=full)
        assert 'https://data.sdss.org/sas/mangawork/manga/spectro/redux/' in url

    @pytest.mark.parametrize('method', [('one'), ('random')])
    def test_onerandom(self, path, method):
        full = self.full(path)
        meth = path.__getattribute__(method)
        one = meth('', full=full)
        if one:
            data = one if method == 'one' else one[0]
            assert re.search(r'(.*?)manga-8485-(\d+)-LOGCUBE(.*?)', data)

    def test_expand(self, path):
        full = self.full(path)
        ex = path.expand('', full=full)
        if ex:
            assert len(ex) > 1

    def test_any(self, path):
        full = self.full(path)
        a = path.any('', full=full)
        if a:
            assert a is True

    def test_refine(self, path):
        full = self.full(path)
        n = path.expand('', full=full)
        if not n:
            return
        items = path.refine(n, r'(.*?)-19\d{2}-(.*?)')
        for item in items:
            assert re.search('8485-190[1-2]', item)

    @pytest.mark.parametrize('copydata',
                             [('mangawork/manga/spectro/redux/v2_4_3/8485/stack/manga-8485-1901-LOGCUBE.fits.gz')],
                             indirect=True, ids=['data'])
    def test_uncompress(self, copydata, monkeysas, path):
        ''' test to find unzipped files with zipped path templates '''
        assert path.templates['mangacube'].endswith('.gz')
        assert path.templates['mangacube'].count('.gz') == 1
        with gzuncompress(copydata) as f:
            full = path.full('mangacube', drpver='v2_4_3', plate=8485, ifu=1901, wave='LOG')
            assert not full.endswith('.gz')
            assert full.count('.gz') == 0
            assert full.endswith('.fits')

    @pytest.mark.parametrize('copydata',
                             [('mangawork/manga/spectro/redux/v2_5_3/8485/images/1901.png')],
                             indirect=True, ids=['data'])
    def test_compress(self, copydata, monkeysas, path):
        ''' test to find zipped files with non-zipped path templates '''
        assert not path.templates['mangaimage'].endswith('.gz')
        assert path.templates['mangaimage'].count('.gz') == 0
        with gzcompress(copydata) as f:
            full = path.full('mangaimage', drpver='v2_5_3', plate=8485, ifu=1901)
            assert not full.endswith('.png')
            assert full.endswith('.gz')
            assert full.count('.gz') == 1

    def test_uncompress_nofileexists(self, monkeysas, path):
        ''' test if no file exists, full returns original template path '''
        assert path.templates['mangacube'].endswith('.gz')
        full = path.full('mangacube', drpver='v2_4_3', plate=8888, ifu=12345, wave='LOG')
        assert full.endswith('.gz')
        assert full.count('.gz') == 1

    @pytest.mark.parametrize('copymulti',
                             [('mangawork/manga/spectro/redux/v2_4_3/8485/stack/manga-8485-*-LOGCUBE.fits.gz')],
                             indirect=True, ids=['data'])
    @pytest.mark.parametrize('plate, ifu', [(8888, '*'), (8888, 12345),
                                            (8485, 1901), (8485, '*')],
                             ids=['nodata-wild', 'nodata', 'glob', 'glob-wild'])
    def test_compression_wildcards(self, copymulti, monkeysas, path, plate, ifu):
        assert path.templates['mangacube'].endswith('.gz')
        full = path.full('mangacube', drpver='v2_4_3', plate=plate, ifu=ifu, wave='LOG')
        assert full.endswith('.gz')
        assert full.count('.gz') == 1

    @pytest.mark.parametrize('mirror', [(True), (False)])
    def test_netloc(self, mirror):
        ''' test the net location and remote_base '''
        path = Path(mirror=mirror)
        if mirror:
            assert path.netloc == 'data.mirror.sdss.org'
            assert path.remote_base == 'https://data.mirror.sdss.org'
        else:
            assert path.netloc == 'data.sdss.org'
            assert path.remote_base == 'https://data.sdss.org'

    def test_path_versions(self, path):
        ff = path.full('mangaimage', plate=8485, ifu=1901, drpver='v2_4_3')
        assert 'stack' not in ff
        assert path.templates['mangaimage'] == '$MANGA_SPECTRO_REDUX/{drpver}/{plate}/images/{ifu}.png'

        path = Path(release='DR13')
        ff = path.full('mangaimage', plate=8485, ifu=1901, drpver='v2_4_3', dir3d='stack')
        assert 'stack' in ff
        assert path.templates['mangaimage'] == '$MANGA_SPECTRO_REDUX/{drpver}/{plate}/{dir3d}/images/{ifu}.png'

    def test_svn_paths(self, path):
        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'mangapreim/trunk/data' in ff
        assert not ff.startswith(os.getenv("SAS_BASE_DIR"))
        assert ff.startswith(os.getenv("PRODUCT_ROOT"))
        loc = path.location('', full=ff)
        assert loc.startswith('data')

    @pytest.mark.parametrize('dr', [(True), (False)])
    def test_svn_urls(self, dr):
        if dr:
            path = Path(release='DR15')
        else:
            path = Path()
        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        url = path.url('', full=ff)
        assert url.startswith('https://svn.sdss.org/')
        if dr:
            assert 'svn.sdss.org/public' in url

    def test_svn_tags(self, path):
        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'mangapreim/trunk/data' in ff

        path = Path(release='DR15')
        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'mangapreim/v2_5/data' in ff

        url = path.url('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'mangapreim/tags/v2_5/data' in url

    def test_svn_url_tag(self):
        path = Path(release='DR15')
        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'mangapreim/v2_5/data' in ff

        url1 = path.url('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        url2 = path.url('', full=ff)

        assert 'mangapreim/tags/v2_5/data' in url1
        assert 'mangapreim/tags/v2_5/data' in url2
        assert url1 == url2

    def test_svn_force_module(self, monkeyoos, path):

        path = Path(release='DR15')
        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'mangapreim/v2_5/data' in ff

        ff = path.full('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007', force_module=True)
        assert 'mangapreim/trunk/data' in ff

    @pytest.mark.parametrize('tree_ver', [('sdsswork'), ('dr15'), ('sdss5'), ('mpl8')])
    def test_release_from_module(self, monkeypatch, tree_ver):
        monkeypatch.setenv('TREE_VER', tree_ver)
        path = Path()
        assert path.release == tree_ver

    def test_sdss5_paths(self, monkeypatch):
        path = Path(release='sdss5')
        assert 'rsFields' in path.templates
        f1 = path.full('rsFields', plan='001', observatory='APO')

        monkeypatch.setenv('TREE_VER', 'sdss5')
        path = Path()
        assert 'rsFields' in path.templates
        f2 = path.full('rsFields', plan='001', observatory='APO')

        assert f1 == f2


@pytest.fixture()
def monkeyoos(monkeypatch):
    ''' monkeypatch the original os environ from tree '''
    oos = tree.get_orig_os_environ()
    monkeypatch.setitem(oos, "MANGAPREIM_DIR", '/tmpdir/data/manga/mangapreim/trunk')
    yield oos
    monkeypatch.delitem(oos, "MANGAPREIM_DIR", raising=False)



from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import input
from sdss_access import is_posix
from sdss_access.path import AccessError
from os.path import join
from os import environ


class Auth(object):
    ''' class for setting up SAS authenticaton for SDSS users '''

    def __init__(self, netloc=None, public=False, verbose=False):
        self.public = public
        self.verbose = verbose
        self.username = None
        self.password = None
        self.set_netloc(netloc=netloc)
        self.set_netrc()

    def set_netrc(self):
        """ add the following username and password to your ~/.netrc file
            and remember to chmod 600 ~/.netrc

            machine data.sdss.org
            login sdss
            password ***-******

            Windows: recommending _netrc following
            https://stackoverflow.com/questions/6031214/git-how-to-use-netrc-file-on-windows-to-save-user-and-password
        """
        # set blank netrc
        self.netrc = None

        # if public do nothing
        if self.public:
            return

        # try to get the netrc file
        try:
            from netrc import netrc
        except Exception as e:
            netrc = None
            if self.verbose:
                print("SDSS_ACCESS> AUTH NETRC: {0}".format(e))

        if netrc:
            netfile = join(environ['HOME'], "_netrc") if not is_posix else None
            try:
                self.netrc = netrc(netfile)
            except FileNotFoundError as e:
                raise AccessError("No netrc file found. Please create one. {0}".format(e))

    def set_netloc(self, netloc=None):
        ''' sets the url domain location '''
        self.netloc = netloc

    def set_username(self, username=None, inquire=None):
        ''' sets the authentication username

        Parameters:
            username: str
                The username for SDSS SAS access
            inquire: bool
                If True, prompts for input of username.
        '''
        self.username = input("user [sdss]: ") or "sdss" if inquire else username

    def set_password(self, password=None, inquire=None):
        ''' sets the authentication password

        Parameters:
            password: str
                The password for SDSS SAS access
            inquire: bool
                If True, prompts for input of password.
        '''
        try:
            from getpass import getpass
            self.password = getpass("password: ") if inquire else password
        except Exception as e:
            if self.verbose:
                print("SDSS_ACCESS> AUTH PASSWORD: {0}".format(e))
            self.password = None

    def ready(self):
        return self.username and self.password

    def load(self):
        ''' Sets the username and password from the local netrc file '''
        if self.netloc and self.netrc:
            authenticators = self.netrc.authenticators(self.netloc)
            if authenticators and len(authenticators) == 3:
                self.set_username(authenticators[0])
                self.set_password(authenticators[2])
                if self.verbose:
                    print("authentication for netloc={0} set for username={1}".format(self.netloc, self.username))
            else:
                if self.verbose:
                    print("cannot find {0} in ~/.netrc".format(self.netloc))
                self.set_username()
                self.set_password()
        else:
            self.set_username()
            self.set_password()


class AuthMixin(object):
    ''' Mixin class to provide authentication method to other classes '''

    def set_auth(self, username=None, password=None, inquire=True):
        ''' Set the authentication

        Parameters:
            username: str
                The username for SDSS SAS access
            password: str
                The password for SDSS SAS access
            inquire: bool
                If True, prompts for input of username/password.
        '''
        self.auth = Auth(public=self.public, netloc=self.netloc, verbose=self.verbose)
        self.auth.set_username(username)
        self.auth.set_password(password)

        # if public then exit
        if self.public:
            return

        # try to load the username and password
        if not self.auth.ready():
            self.auth.load()

        # if still not ready then prompt for username and password
        if not self.auth.ready():
            self.auth.set_username(inquire=inquire)
            self.auth.set_password(inquire=inquire)
